package meerkat.bulletinboard;

import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;

import java.util.List;

/**
 * Created by Arbel on 07/11/15.
 * 
 * This interface refers to a single instance of a Bulletin Board
 * An implementation of this interface may use any DB and be hosted on any machine.
 */

public interface BulletinBoardServer{
	
	/**
	 * This method initializes the server by reading the signature data and storing it
	 * It also establishes the connection to the DB
	 * @throws CommunicationException on DB connection error
	 */
	public void init(String meerkatDB) throws CommunicationException;
	
	/**
     * Post a message to bulletin board.
     * @param msg is the actual (signed) message
	 * @return TRUE if the message has been authenticated and FALSE otherwise (in ProtoBuf form)
	 * @throws CommunicationException on DB connection error
     */
	public BoolMsg postMessage(BulletinBoardMessage msg) throws CommunicationException;
	 
	 /**
	  * Read all messages posted matching the given filter
	  * @param filterList return only messages that match the filters (empty list or null means no filtering)
	  * @return
	  * @throws CommunicationException on DB connection error
	  */
	public BulletinBoardMessageList readMessages(MessageFilterList filterList) throws CommunicationException;

	/**
	 * Informs server about a new batch message
	 * @param message contains the required data about the new batch
	 * @return TRUE if the batch request is accepted amd FALSE otherwise
	 * 		Specifically, if such a batch already exists and is not yet closed: the value returned will be TRUE
	 * 		However, if such a batch exists and is already closed: the value returned will be FALSE
	 * @throws CommunicationException on DB connection error
     */
	public BoolMsg beginBatch(BeginBatchMessage message) throws CommunicationException;

	/**
	 * Posts a (part of a) batch message to the bulletin board
	 * Note that the existence and contents of a batch message are not available for reading before the batch is finalized
	 * @param batchMessage contains the (partial) data this message carries as well as meta-data required in order to place the data
	 *                     in the correct position inside the correct batch
	 * @return TRUE if the message is accepted and successfully saved and FALSE otherwise
	 * 		Specifically, if the batch is already closed: the value returned will be FALSE
	 * 		However, requiring to open a batch before insertion of messages is implementation-dependent
	 * @throws CommunicationException on DB connection error
     */
	public BoolMsg postBatchMessage(BatchMessage batchMessage) throws CommunicationException;

	/**
	 * Attempts to close and finalize a batch message
	 * @param message contains the data necessary to close the batch; in particular: the signature for the batch
	 * @return TRUE if the batch was successfully closed, FALSE otherwise
	 * 		Specifically, if the signature is invalid or if some of the batch parts have not yet been submitted: the value returned will be FALSE
	 * @throws CommunicationException on DB connection error
     */
	public BoolMsg closeBatchMessage(CloseBatchMessage message) throws CommunicationException;

	/**
	 * Reads a batch message from the server (starting with the supplied position)
	 * @param message specifies the signer ID and the batch ID to read as well as an (optional) start position
	 * @return an ordered list of batch messages starting from the specified start position (if given) or from the beginning (if omitted)
	 * @throws CommunicationException on DB connection error
	 * @throws IllegalArgumentException if message does not specify a batch
     */
	public List<BatchData> readBatch(BatchSpecificationMessage message) throws CommunicationException, IllegalArgumentException;

	 /**
	 * This method closes the connection to the DB
	 * @throws CommunicationException on DB connection error
	 */
	public void close() throws CommunicationException;
}
